"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

# pylint:disable=duplicate-code

from unittest import IsolatedAsyncioTestCase, mock

import pytest
from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ChannelAccount, ConversationAccount

from teams import Application
from tests.utils import SimpleAdapter


class TestMeetings(IsolatedAsyncioTestCase):
    app: Application

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.app = Application()
        yield

    @pytest.mark.asyncio
    async def test_start(self):
        handler = mock.AsyncMock()
        self.app.meetings.start()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingStart",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_start_with_invalid_name(self):
        handler = mock.AsyncMock()
        self.app.meetings.start()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meeting",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_start_with_missing_value(self):
        handler = mock.AsyncMock()
        self.app.meetings.start()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingStart",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_start_with_wrong_type(self):
        handler = mock.AsyncMock()
        self.app.meetings.start()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    name="application/vnd.microsoft.meetingStart",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_start_with_wrong_channel(self):
        handler = mock.AsyncMock()
        self.app.meetings.start()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingStart",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="team",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_end(self):
        handler = mock.AsyncMock()
        self.app.meetings.end()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingEnd",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_end_with_missing_value(self):
        handler = mock.AsyncMock()
        self.app.meetings.end()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingEnd",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_end_with_invalid_name(self):
        handler = mock.AsyncMock()
        self.app.meetings.end()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meeting",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_end_with_wrong_type(self):
        handler = mock.AsyncMock()
        self.app.meetings.end()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    name="application/vnd.microsoft.meetingEnd",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_end_with_wrong_channel(self):
        handler = mock.AsyncMock()
        self.app.meetings.end()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    name="application/vnd.microsoft.meetingEnd",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="team",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "meetingType": "scheduled",
                        "startTime": "2020-10-01T00:00:00.000Z",
                        "endTime": "2020-10-01T00:30:00.000Z",
                    },
                ),
            )
        )

        handler.assert_not_called()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class AdaptiveCardsSearchResult(dict):
    title: str
    value: str

    def __init__(self, title, value):
        dict.__init__(self, title=title, value=value)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional, Union

from dataclasses_json import DataClassJsonMixin, dataclass_json

from ..prompts.message import Message


@dataclass_json
@dataclass
class PredictedDoCommand(DataClassJsonMixin):
    """
    A predicted DO command is an action that the AI system should perform.
    """

    type: Literal["DO"] = "DO"
    "Type to indicate that a DO command is being returned."

    action: str = ""
    "The named action that the AI system should perform."

    parameters: Dict[str, Any] = field(default_factory=dict)
    "Any parameters that the AI system should use to perform the action."


@dataclass_json
@dataclass
class PredictedSayCommand(DataClassJsonMixin):
    """
    A predicted SAY command is a response that the AI system should say.
    """

    type: Literal["SAY"] = "SAY"
    "Type to indicate that a SAY command is being returned."

    response: Optional[Message[str]] = None
    "The prompt response containing what the AI system should say."


PredictedCommand = Union[PredictedDoCommand, PredictedSayCommand]
"A predicted command is a command that the AI system should execute."


@dataclass
class Plan:
    """
    A plan is a set of commands that the AI system will execute.
    """

    type: Literal["plan"] = "plan"
    "Type to indicate that a plan is being returned."

    commands: List[PredictedCommand] = field(default_factory=list)
    "Array of predicted commands that the AI system should execute."

    @classmethod
    def from_dict(cls, dict: Dict[str, Any], *, infer_missing: bool = False) -> "Plan":
        commands: List[PredictedCommand] = []
        for command in dict["commands"]:
            if command["type"] == "DO":
                commands.append(PredictedDoCommand.from_dict(command, infer_missing=infer_missing))
            if command["type"] == "SAY":
                commands.append(PredictedSayCommand.from_dict(command, infer_missing=infer_missing))
        return Plan(commands=commands)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Generic, Optional, TypeVar, cast

from botbuilder.core import TurnContext
from botbuilder.dialogs import (
    ComponentDialog,
    DialogContext,
    DialogTurnResult,
    DialogTurnStatus,
    PromptOptions,
)

from ..state import ConversationState, TurnState

StateT = TypeVar("StateT", bound=TurnState)


class Dialog(ComponentDialog, Generic[StateT]):
    "base class for dialog implementations"

    async def create_dialog_context(self, context: TurnContext, state: StateT) -> DialogContext:
        """
        Create the dialog context

        Args:
            context (TurnContext): the turn context
            state (StateT): the turn state
            property (str): the turn state property

        Returns:
            The dialog context
        """
        self._dialogs._dialog_state = cast(ConversationState, state.conversation).create_property(
            self.id
        )
        return await self._dialogs.create_context(context)

    async def run_dialog(
        self, context: TurnContext, state: StateT, *, options: Optional[PromptOptions] = None
    ) -> DialogTurnResult:
        """
        Run or continue the authentication dialog.

        Args:
            context (TurnContext): the turn context
            state (StateT): the turn state
            property (str): the turn state property

        Returns:
            Dialog turn result that contains token if sign in successs
        """
        ctx = await self.create_dialog_context(context, state)
        res = await ctx.continue_dialog()

        if res.status == DialogTurnStatus.Empty:
            res = await ctx.begin_dialog(self.initial_dialog_id or "default", options)

        return res



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.prompt_response import PromptResponse
from ..tokenizers import Tokenizer
from .validation import Validation


class PromptResponseValidator(ABC):
    """
    A validator that can be used to validate prompt responses.
    """

    @abstractmethod
    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse,
        remaining_attempts: int,
    ) -> Validation:
        """
        Validates a response to a prompt.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            response (PromptResponse): response Response to validate.
            remaining_attempts (int): Number of remaining attempts to validate the response.

        Returns:
            Validation: A `Validation` object.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from teams.state import State


class TestState(IsolatedAsyncioTestCase):
    async def test_should_set_attribute(self):
        state = await State.load(MagicMock())
        self.assertFalse("test" in state)
        state.test = 1
        self.assertTrue("test" in state)
        self.assertEqual(state.test, 1)
        self.assertEqual(state["test"], 1)

    async def test_should_set_item(self):
        state = await State.load(MagicMock())
        self.assertFalse("test" in state)
        state["test"] = 1
        self.assertTrue("test" in state)
        self.assertEqual(state["test"], 1)
        self.assertEqual(state.test, 1)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import re
from typing import List, Optional

from teams.ai.citations import ClientCitation


def get_used_citations(
    text: str, citations: List[ClientCitation]
) -> Optional[List[ClientCitation]]:
    """
    Get the citations used in the text. This will remove any citations that are
    included in the citations array from the response but not referenced in the text.
    Args:
        text: str The text to search for citations.
        citations: List[ClientCitation] The list of citations to search for.
    Returns:
        Optional[List[ClientCitation]]: The list of citations used in the text.
    """
    regex = r"\[(\d+)\]"
    matches = re.findall(regex, text)

    if not matches:
        return None

    used_citations = []
    for match in matches:
        for citation in citations:
            if citation.position == match:
                used_citations.append(citation)
                break
    return used_citations



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, List

from .parse_object import parse_object


def parse(text: str) -> List[Any]:
    """
    Parse all objects from a response string.
    """

    objects: List[Any] = []
    lines = text.split("\n")

    for line in lines:
        obj = parse_object(line)

        if obj is not None and len(obj) > 0:
            objects.append(obj)

    if len(objects) == 0:
        obj = parse_object(text)

        if obj is not None:
            objects.append(obj)

    return objects



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import base64
from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import TurnContext

from teams import InputFile
from teams.ai.prompts import (
    ImageContentPart,
    ImageUrl,
    PromptFunctions,
    TextContentPart,
    UserInputMessage,
)
from teams.ai.tokenizers import GPTTokenizer
from teams.state import ConversationState, TempState, TurnState, UserState


class TestUserInputMessage(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.context = MagicMock(spec=TurnContext)
        self.memory = await TurnState[ConversationState, UserState, TempState].load(self.context)
        self.functions = MagicMock(spec=PromptFunctions)
        self.tokenizer = GPTTokenizer()
        await self.memory.load(self.context)

    def test_init_default(self):
        message = UserInputMessage()
        self.assertEqual(message.tokens, -1)
        self.assertEqual(message._input_variable, "input")
        self.assertEqual(message._files_variable, "input_files")

    def test_init_with_optional_params(self):
        message = UserInputMessage(input_variable="customInput", files_variable="customFiles")
        self.assertEqual(message._input_variable, "customInput")
        self.assertEqual(message._files_variable, "customFiles")

    async def test_render_as_messages_no_input(self):
        message = UserInputMessage()
        del self.memory.temp
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.length, 0)
        self.assertEqual(len(result.output), 1)
        self.assertEqual(result.output[0].content, [])
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_text(self):
        message = UserInputMessage()
        self.memory.temp.input = "Hello, world!"
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.length, 4)
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 1)
        self.assertTrue(isinstance(result.output[0].content[0], TextContentPart))
        self.assertEqual(result.output[0].content[0].text, "Hello, world!")
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_text_truncated(self):
        message = UserInputMessage()
        self.memory.temp.input = "Hello, world!"
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 2
        )
        self.assertEqual(result.length, 2)
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 1)
        self.assertTrue(isinstance(result.output[0].content[0], TextContentPart))
        self.assertEqual(result.output[0].content[0].text, "Hello,")
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_files(self):
        message = UserInputMessage()
        base64_string = "SGVsbG8gd29ybGQ="  # Mock the expected content
        self.memory.temp.input_files = [
            InputFile(base64.b64decode(base64_string), "image/png", None)
        ]
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.length, 85)
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 1)
        self.assertIsInstance(result.output[0].content[0], ImageContentPart)
        self.assertIsInstance(result.output[0].content[0].image_url, ImageUrl)
        self.assertEqual(
            result.output[0].content[0].image_url.url, f"data:image/png;base64,{base64_string}"
        )
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_files_truncated(self):
        message = UserInputMessage()
        base64_string = "SGVsbG8gd29ybGQ="  # Mock the expected content
        self.memory.temp.input_files = [
            InputFile(base64.b64decode(base64_string), "image/png", None)
        ]
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 50
        )
        self.assertEqual(result.length, 0)
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 0)
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_text_and_files(self):
        message = UserInputMessage()
        self.memory.temp.input = "Hello, world!"
        base64_string = "SGVsbG8gd29ybGQ="  # Mock the expected content
        self.memory.temp.input_files = [
            InputFile(base64.b64decode(base64_string), "image/png", None)
        ]
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 100
        )
        self.assertEqual(result.length, 4 + 85)  # Length of text and image
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 2)
        self.assertTrue(isinstance(result.output[0].content[0], TextContentPart))
        self.assertEqual(result.output[0].content[0].text, "Hello, world!")
        self.assertIsInstance(result.output[0].content[1], ImageContentPart)
        self.assertIsInstance(result.output[0].content[1].image_url, ImageUrl)
        self.assertEqual(
            result.output[0].content[1].image_url.url, f"data:image/png;base64,{base64_string}"
        )
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)

    async def test_render_as_messages_with_text_and_files_truncated(self):
        message = UserInputMessage()
        self.memory.temp.input = "Hello, world!"
        base64_string = "SGVsbG8gd29ybGQ="  # Mock the expected content
        self.memory.temp.input_files = [
            InputFile(base64.b64decode(base64_string), "image/png", None)
        ]
        result = await message.render_as_messages(
            self.context, self.memory, self.functions, self.tokenizer, 85
        )
        self.assertEqual(result.length, 4)
        self.assertEqual(len(result.output), 1)
        assert result.output[0].content is not None
        self.assertEqual(len(result.output[0].content), 1)
        self.assertTrue(isinstance(result.output[0].content[0], TextContentPart))
        self.assertEqual(result.output[0].content[0].text, "Hello, world!")
        self.assertEqual(result.output[0].role, "user")
        self.assertFalse(result.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from dataclasses import dataclass, field
from logging import Logger
from typing import Any, List, Optional

from botbuilder.core import TurnContext

from ...state import Memory, MemoryBase
from ..models import PromptCompletionModel, PromptResponse
from ..prompts import (
    ConversationHistorySection,
    Message,
    Prompt,
    PromptFunctions,
    PromptTemplate,
)
from ..tokenizers import Tokenizer
from ..validators import DefaultResponseValidator, PromptResponseValidator


@dataclass
class LLMClientOptions:
    """
    Options for an LLMClient instance.
    """

    model: PromptCompletionModel
    "AI model to use for completing prompts."

    history_variable: str = "conversation.history"
    """
    Optional. Memory variable used for storing conversation history.
    Defaults to `conversation.history`
    """

    input_variable: str = "temp.input"
    """
    Optional. Memory variable used for storing the users input message.
    Defaults to `temp.input`
    """

    max_history_messages: int = 10
    """
    Optional. Maximum number of conversation history messages to maintain.
    Defaults to `10`
    """

    max_repair_attempts: int = 3
    """
    Optional. Maximum number of automatic repair attempts the LLMClient instance will make.
    Defaults to `3`
    """

    validator: PromptResponseValidator = field(default_factory=DefaultResponseValidator)
    """
    Optional. Response validator to use when completing prompts.
    Defaults to `DefaultResponseValidator`
    """

    logger: Optional[Logger] = None
    """
    Optional. When set the model will log requests
    """


class LLMClient:
    """
    LLMClient class that's used to complete prompts.
    """

    _options: LLMClientOptions

    @property
    def options(self) -> LLMClientOptions:
        return self._options

    def __init__(self, options: LLMClientOptions) -> None:
        """
        Creates a new `LLMClient` instance.

        Args:
            options (LLMClientOptions): Options to configure the instance with.
        """

        self._options = options

    def add_function_result_to_history(self, memory: MemoryBase, name: str, results: Any) -> None:
        """
        Adds a result from a `function_call` to the history.

        Args:
            memory (MemoryBase): An interface for accessing state values.
            name (str): Name of the function that was called.
            results (Any): Results returned by the function.
        """

        content = ""

        if isinstance(results, object):
            content = json.dumps(results)
        else:
            content = str(results)

        self._add_message_to_history(
            memory=memory,
            variable=self._options.history_variable,
            message=Message(role="function", name=name, content=content),
        )

    async def complete_prompt(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        template: PromptTemplate,
        remaining_attempts: Optional[int] = None,
    ) -> PromptResponse[str]:
        """
        Completes a prompt.

        Args:
            content (TurnContext): The turn context.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Functions to use when rendering the prompt.
            tokenizer (Tokenizer): Tokenizer used when rendering the prompt or counting tokens.
            template (PromptTemplate): Prompt used for the conversation.
        """

        remaining_attempts = remaining_attempts or self._options.max_repair_attempts

        try:
            if remaining_attempts <= 0:
                return PromptResponse(
                    status="invalid_response", error="Reached max model response repair attempts."
                )

            res = await self._options.model.complete_prompt(
                context=context,
                memory=memory,
                functions=functions,
                tokenizer=tokenizer,
                template=template,
            )

            if res.status != "success" or not res.message:
                return res

            if not res.input:
                res.input = Message(role="user", content=memory.get(self._options.input_variable))

            validation = await self._options.validator.validate_response(
                context=context,
                memory=memory,
                tokenizer=tokenizer,
                response=res,
                remaining_attempts=remaining_attempts,
            )

            if validation.value:
                res.message.content = validation.value

            if not validation.valid:
                fork = Memory(memory)

                if self._options.logger:
                    self._options.logger.info(f"REPAIRING RESPONSE:\n{res.message.content or ''}")

                self._add_message_to_history(
                    fork, f"{self._options.history_variable}-repair", res.message
                )

                self._add_message_to_history(
                    fork,
                    f"{self._options.history_variable}-repair",
                    Message(
                        role="user",
                        content=validation.feedback
                        or "The response was invalid. Try another strategy.",
                    ),
                )

                return await self.complete_prompt(
                    context=context,
                    memory=fork,
                    functions=functions,
                    tokenizer=tokenizer,
                    template=PromptTemplate(
                        name=template.name,
                        actions=template.actions,
                        augmentation=template.augmentation,
                        config=template.config,
                        prompt=Prompt(
                            sections=[
                                template.prompt,
                                ConversationHistorySection(
                                    variable=f"{self._options.history_variable}-repair"
                                ),
                            ]
                        ),
                    ),
                    remaining_attempts=remaining_attempts - 1,
                )

            self._add_message_to_history(memory, self._options.history_variable, res.input)
            self._add_message_to_history(memory, self._options.history_variable, res.message)
            return res
        except Exception as err:  # pylint: disable=broad-except
            return PromptResponse(status="error", error=str(err))

    def _add_message_to_history(
        self, memory: MemoryBase, variable: str, message: Message[Any]
    ) -> None:
        history: List[Message] = memory.get(variable) or []
        history.append(message)

        if len(history) > self._options.max_history_messages:
            del history[0 : len(history) - self._options.max_history_messages]

        memory.set(variable, history)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

# pylint:disable=duplicate-code

from unittest import IsolatedAsyncioTestCase, mock

import pytest
from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ChannelAccount, ConversationAccount
from botbuilder.schema.teams import (
    MessagingExtensionAction,
    MessagingExtensionResult,
    TaskModuleTaskInfo,
)

from teams import Application
from tests.utils import SimpleAdapter


class TestMessageExtensions(IsolatedAsyncioTestCase):
    app: Application

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.app = Application()
        yield

    @pytest.mark.asyncio
    async def test_query(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.query("256")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/query",
                    text="test",
                    value={"commandId": "256"},
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_query_link(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.query_link("256")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/queryLink",
                    text="test",
                    value={"commandId": "256", "url": ""},
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_anonymous_query_link(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.anonymous_query_link("256")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/anonymousQueryLink",
                    text="test",
                    value={"commandId": "256", "url": ""},
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_message_preview(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.message_preview("256", "edit")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/submitAction",
                    text="test",
                    value=MessagingExtensionAction(
                        command_id="256",
                        bot_message_preview_action="edit",
                        bot_activity_preview=[Activity()],
                    ),
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_fetch_task(self):
        handler = mock.AsyncMock()
        handler.return_value = TaskModuleTaskInfo()
        self.app.message_extensions.fetch_task("256")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/fetchTask",
                    text="test",
                    value=MessagingExtensionAction(command_id="256"),
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_select_item(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.select_item()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/selectItem",
                    text="test",
                    value="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_submit_action(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.submit_action("256")(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/submitAction",
                    text="test",
                    value=MessagingExtensionAction(command_id="256"),
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_query_setting_url(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.query_setting_url()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/querySettingUrl",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_query_setting_url_invalid_type(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.query_setting_url()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="composeExtension/querySettingUrl",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_query_setting_url_invalid_name(self):
        handler = mock.AsyncMock()
        handler.return_value = MessagingExtensionResult()
        self.app.message_extensions.query_setting_url()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/queryUrl",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_configure_settings(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.configure_settings()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/setting",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"theme": "dark"},
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_configure_settings_empty_value(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.configure_settings()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/setting",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_configure_settings_wrong_type(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.configure_settings()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="composeExtension/setting",
                    text="test",
                    value={"theme": "dark"},
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_configure_settings_wrong_name(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.configure_settings()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/settings",
                    text="test",
                    value={"theme": "dark"},
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_card_button_clicked(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.card_button_clicked()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/onCardButtonClicked",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"title": "Query button", "display_text": "Yes", "value": "Yes"},
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_card_button_clicked_empty_value(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.card_button_clicked()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/onCardButtonClicked",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        handler.assert_called_once()

    @pytest.mark.asyncio
    async def test_card_button_clicked_wrong_type(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.card_button_clicked()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="composeExtension/onCardButtonClicked",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"title": "Query button", "display_text": "Yes", "value": "Yes"},
                ),
            )
        )

        handler.assert_not_called()

    @pytest.mark.asyncio
    async def test_card_button_clicked_invalid_name(self):
        handler = mock.AsyncMock()
        self.app.message_extensions.card_button_clicked()(handler)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="composeExtension/onCardButtonClick",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="msteams",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"title": "Query button", "display_text": "Yes", "value": "Yes"},
                ),
            )
        )

        handler.assert_not_called()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .action_augmentation_section import ActionAugmentationSection
from .conversation_history_section import ConversationHistorySection
from .data_source_section import DataSourceSection
from .group_section import GroupSection
from .layout_engine_section import LayoutEngineSection
from .prompt_section import PromptSection
from .prompt_section_base import PromptSectionBase
from .template_section import TemplateSection
from .text_section import TextSection

__all__ = [
    "ActionAugmentationSection",
    "ConversationHistorySection",
    "DataSourceSection",
    "GroupSection",
    "LayoutEngineSection",
    "PromptSection",
    "PromptSectionBase",
    "TemplateSection",
    "TextSection",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from enum import Enum
from typing import Awaitable, Callable, List

from botbuilder.core import TurnContext

from ....app_error import ApplicationError
from ....state import MemoryBase
from ....utils.to_string import to_string
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section_base import PromptSectionBase


# private
class _ParseState(Enum):
    IN_TEXT = 1
    IN_PARAMETER = 2
    IN_STRING = 3


# private
_PartRenderer = Callable[[TurnContext, MemoryBase, PromptFunctions, Tokenizer, int], Awaitable[str]]


class TemplateSection(PromptSectionBase):
    """
    A template section that will be rendered as a message.

    This section type is used to render a template as a message. The template can contain
    parameters that will be replaced with values from memory or call functions to generate
    dynamic content.

    Template syntax:
    - `{{$memoryKey}}` - Renders the value of the specified memory key.
    - `{{functionName}}` - Calls the specified function and renders the result.
    - `{{functionName arg1 arg2 ...}}` - Calls the specified function with the
      provided list of arguments.

    Function arguments are optional and separated by spaces.
    They can be quoted using ', ", or ` delimiters.
    """

    _template: str
    _role: str
    _parts: List[_PartRenderer]

    # pylint: disable=too-many-arguments # No argument can be removed based on the design
    def __init__(
        self,
        template: str,
        role: str,
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n",
        text_prefix: str = "",
    ):
        """
        Creates a new 'TemplateSection' instance.

        Args:
            template (str): Template to use for this section.
            role (str): Message role to use for this section.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
            required (bool, optional): Indicates if this section is required. Defaults to `True`.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n`.
            text_prefix (str, optional): The text prefix. Prefix to use for text output.
              Defaults to ''.

        """
        super().__init__(tokens, required, separator, text_prefix)
        self._template = template
        self._role = role
        self._parts: List[_PartRenderer] = []
        self._parse_template()

    # pylint: enable=too-many-arguments

    @property
    def template(self) -> str:
        """str: The template string."""
        return self._template

    @property
    def role(self) -> str:
        """str: The role of the template."""
        return self._role

    # pylint: disable=too-many-arguments # No argument can be removed based on the design
    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """

        rendered_parts = [
            await part(context, memory, functions, tokenizer, max_tokens) for part in self._parts
        ]

        # Join all parts
        text = "".join(rendered_parts)
        length = len(tokenizer.encode(text))

        # Return output
        messages = [Message(self.role, text)] if length > 0 else []
        return self._return_messages(messages, length, tokenizer, max_tokens)

    # pylint: enable=too-many-arguments

    def _parse_template(self):
        # Parse template
        part = ""
        state = _ParseState.IN_TEXT
        string_delim = ""
        i = 0
        while i < len(self.template):
            char = self.template[i]
            if state == _ParseState.IN_TEXT:
                if char == "{" and i + 1 < len(self.template) and self.template[i + 1] == "{":
                    if len(part) > 0:
                        self._parts.append(self._create_text_renderer(part))
                        part = ""
                    state = _ParseState.IN_PARAMETER
                    i += 1
                else:
                    part += char
            elif state == _ParseState.IN_PARAMETER:
                if char == "}" and i + 1 < len(self.template) and self.template[i + 1] == "}":
                    if len(part) > 0:
                        part = part.strip()
                        if part[0] == "$":
                            self._parts.append(self._create_variable_renderer(part[1:]))
                        else:
                            self._parts.append(self._create_function_renderer(part))
                        part = ""

                    state = _ParseState.IN_TEXT
                    i += 1
                elif char in ["'", '"', "`"]:
                    string_delim = char
                    state = _ParseState.IN_STRING
                    part += char
                else:
                    part += char
            elif state == _ParseState.IN_STRING:
                part += char
                if char == string_delim:
                    state = _ParseState.IN_PARAMETER

            i += 1

        # Ensure we ended in the correct state
        if state != _ParseState.IN_TEXT:
            raise ApplicationError(f"Invalid template: {self.template}")

        # Add final part
        if len(part) > 0:
            self._parts.append(self._create_text_renderer(part))

    def _create_text_renderer(self, text: str) -> _PartRenderer:
        async def renderer(
            _context: TurnContext,
            _memory: MemoryBase,
            _functions: PromptFunctions,
            _tokenizer: Tokenizer,
            _max_tokens: int,
        ) -> str:
            return text

        return renderer

    def _create_variable_renderer(self, name: str) -> _PartRenderer:
        async def renderer(
            _context: TurnContext,
            memory: MemoryBase,
            _functions: PromptFunctions,
            tokenizer: Tokenizer,
            _max_tokens: int,
        ) -> str:
            return to_string(tokenizer, memory.get(name))

        return renderer

    def _create_function_renderer(self, param: str) -> _PartRenderer:
        name = ""
        args: List[str] = []

        # Parse function name and args
        part = ""
        state = _ParseState.IN_TEXT
        string_delim = ""
        for _, char in enumerate(param):
            if state == _ParseState.IN_TEXT:
                if char in ["'", '"', "`"]:
                    if len(part) > 0:
                        if not name:
                            name = part
                        else:
                            args.append(part)
                        part = ""
                    string_delim = char
                    state = _ParseState.IN_STRING
                elif char == " ":
                    if len(part) > 0:
                        if not name:
                            name = part
                        else:
                            args.append(part)
                        part = ""
                else:
                    part += char
            elif state == _ParseState.IN_STRING:
                if char == string_delim:
                    if len(part) > 0:
                        if not name:
                            name = part
                        else:
                            args.append(part)
                        part = ""
                    state = _ParseState.IN_TEXT
                else:
                    part += char

        # Add final part
        if len(part) > 0:
            if not name:
                name = part
            else:
                args.append(part)

        async def renderer(
            context: TurnContext,
            memory: MemoryBase,
            functions: PromptFunctions,
            tokenizer: Tokenizer,
            _max_tokens: int,
        ) -> str:
            value = await functions.invoke_function(name, context, memory, tokenizer, args)
            return to_string(tokenizer, value)

        return renderer



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.prompts import LayoutEngineSection, PromptFunctions, TextSection
from teams.ai.tokenizers import GPTTokenizer
from teams.state import TurnState


class TestLayoutEngine(IsolatedAsyncioTestCase):
    async def test_render_as_text_empty_sections(self):
        layout_engine = LayoutEngineSection([], 1, False, "\n")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=1,
        )

        self.assertEqual(result.output, "")
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_text_fixed_sections(self):
        section1 = TextSection("Hello World!", "user")
        section2 = TextSection("Teams-AI", "user")
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=100,
        )

        self.assertEqual(result.output, "Hello World! Teams-AI")
        self.assertEqual(result.length, 6)
        self.assertFalse(result.too_long)

    async def test_render_as_text_fixed_sections_too_long(self):
        section1 = TextSection("Hello World!", "user")
        section2 = TextSection("Teams-AI", "user")
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "Hello World! Teams-AI")
        self.assertEqual(result.length, 6)
        self.assertTrue(result.too_long)

    async def test_render_as_text_fixed_sections_drop_optional(self):
        section1 = TextSection("Hello World!", "user")
        section2 = TextSection("Teams-AI", "user", required=False)
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "Hello World!")
        self.assertEqual(result.length, 3)
        self.assertFalse(result.too_long)

    async def test_render_as_text_fixed_sections_too_long_after_drop(self):
        section1 = TextSection("Hello World!", "user")
        section2 = TextSection("Teams-AI", "user", required=False)
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=2,
        )

        self.assertEqual(result.output, "Hello World!")
        self.assertEqual(result.length, 3)
        self.assertTrue(result.too_long)

    async def test_render_as_text_proportional_sections(self):
        section1 = TextSection("Hello World!", "user", tokens=0.5)
        section2 = TextSection("Teams-AI", "user", tokens=0.5)
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "Hello World! Teams-AI")
        self.assertEqual(result.length, 6)
        self.assertTrue(result.too_long)

    async def test_render_as_text_proportional_sections_drop_optional(self):
        section2 = TextSection("Teams-AI!", "user", tokens=0.5, required=False)
        section1 = TextSection("Hello World!", "user", tokens=0.5, required=False)
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=6,
        )

        self.assertEqual(
            result.output, "Hello World!"
        )  # Teams-AI! should be dropped as Hello World! gets enough token capacity
        self.assertEqual(result.length, 3)
        self.assertFalse(result.too_long)

    async def test_render_as_text_proportional_sections_too_long_after_drop(self):
        section1 = TextSection("Hello World!", "user", tokens=0.5)
        section2 = TextSection("Teams-AI!", "user", tokens=0.5, required=False)
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=2,
        )

        self.assertEqual(result.output, "Hello World!")
        self.assertEqual(result.length, 3)
        self.assertTrue(result.too_long)

    async def test_render_as_messages_empty_sections(self):
        layout_engine = LayoutEngineSection([], 1, False, "\n")
        result = await layout_engine.render_as_messages(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=1,
        )

        self.assertEqual(result.output, [])
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_messages(self):
        section1 = TextSection("Hello World!", "user")
        section2 = TextSection("Teams-AI", "user")
        layout_engine = LayoutEngineSection([section1, section2], 1, False, " ")
        result = await layout_engine.render_as_messages(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=100,
        )

        self.assertEqual(result.output[0].content, "Hello World!")
        self.assertEqual(result.output[1].content, "Teams-AI")
        self.assertEqual(result.length, 6)
        self.assertFalse(result.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock

from botbuilder.core import TurnContext

from teams.ai.prompts import (
    FunctionCall,
    ImageContentPart,
    Message,
    PromptFunctions,
    PromptSectionBase,
    RenderedPromptSection,
    TextContentPart,
)
from teams.ai.tokenizers import GPTTokenizer
from teams.state import TurnState


class TestablePromptSectionBase(PromptSectionBase):
    __test__ = False

    def __init__(
        self,
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n",
        text_prefix: str = "",
    ):
        super().__init__(tokens, required, separator, text_prefix)

    async def render_as_messages(self, context, memory, functions, tokenizer, max_tokens):
        return RenderedPromptSection([], 0, False)


class TestPromptSectionBase(IsolatedAsyncioTestCase):
    async def test_render_as_text_empty_message(self):
        section = TestablePromptSectionBase()
        section.render_as_messages = AsyncMock(return_value=RenderedPromptSection([], 0, False))
        result = await section.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "")
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_text_multiple_messages(self):
        section = TestablePromptSectionBase()
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        section.render_as_messages = AsyncMock(
            return_value=RenderedPromptSection(messages, 6, False)
        )
        result = await section.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=100,
        )

        self.assertEqual(result.output, "Hello World!\nTeams-AI")
        self.assertEqual(result.length, 7)
        self.assertFalse(result.too_long)

    async def test_render_as_text_multiple_messages_too_long(self):
        section = TestablePromptSectionBase()
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        section.render_as_messages = AsyncMock(
            return_value=RenderedPromptSection(messages, 6, False)
        )
        result = await section.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "Hello World!\nTeams-AI")
        self.assertEqual(result.length, 7)
        self.assertTrue(result.too_long)

    async def test_render_as_text_multiple_truncate_fixed_length(self):
        section = TestablePromptSectionBase(tokens=5)
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        section.render_as_messages = AsyncMock(
            return_value=RenderedPromptSection(messages, 6, False)
        )
        result = await section.render_as_text(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            max_tokens=5,
        )

        self.assertEqual(result.output, "Hello World!\nTeams-A")
        self.assertEqual(result.length, 5)
        self.assertFalse(result.too_long)

    async def test_get_token_buge_fixed_token(self):
        section = TestablePromptSectionBase(tokens=5)
        result = section._get_token_budget(10)
        self.assertEqual(result, 5)
        result = section._get_token_budget(3)
        self.assertEqual(result, 3)

    async def test_get_token_buge_proportional_token(self):
        section = TestablePromptSectionBase(tokens=0.5)
        result = section._get_token_budget(10)
        self.assertEqual(result, 10)

    async def test_return_messages_fixed_sections(self):
        section = TestablePromptSectionBase()
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        result = section._return_messages(messages, 6, GPTTokenizer(), 100)
        self.assertEqual(result.output, messages)
        self.assertEqual(result.length, 6)
        self.assertFalse(result.too_long)

    async def test_return_messages_fixed_sections_too_long(self):
        section = TestablePromptSectionBase(tokens=5)
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        result = section._return_messages(messages, 6, GPTTokenizer(), 4)
        self.assertEqual(result.output[0].content, "Hello World!")
        self.assertEqual(result.output[1].content, "Teams-A")
        self.assertEqual(result.length, 5)
        self.assertTrue(result.too_long)

    async def test_return_messages_proportional_sections(self):
        section = TestablePromptSectionBase(tokens=0.5)
        messages = [Message("User", "Hello World!"), Message("User", "Teams-AI")]
        result = section._return_messages(messages, 6, GPTTokenizer(), 4)
        self.assertEqual(result.output[0].content, "Hello World!")
        self.assertEqual(result.output[1].content, "Teams-AI")
        self.assertEqual(result.length, 6)
        self.assertTrue(result.too_long)

    async def test_get_message_text(self):
        result = PromptSectionBase.get_message_text(Message("User", "Hello World!"))
        self.assertEqual(result, "Hello World!")

    async def test_get_message_text_with_name(self):
        result = PromptSectionBase.get_message_text(
            Message("User", "Hello World!", name="TestCase")
        )
        self.assertEqual(result, "TestCase returned Hello World!")

    async def test_get_message_text_content_part_list(self):
        result = PromptSectionBase.get_message_text(
            Message(
                "User",
                [
                    TextContentPart("text", "Hello"),
                    TextContentPart("text", "World!"),
                    ImageContentPart("image_url", "https://www.microsoft.com"),
                ],
            )
        )
        self.assertEqual(result, "Hello World!")

    async def test_get_message_text_function_call(self):
        result = PromptSectionBase.get_message_text(
            Message("User", function_call=FunctionCall("function name", "arguments"))
        )
        self.assertEqual(result, '{"name": "function name", "arguments": "arguments"}')

    def test_required_property(self):
        prompt_section_base = TestablePromptSectionBase(
            tokens=0.5, required=True, separator=";", text_prefix="prefix"
        )
        self.assertEqual(prompt_section_base.required, True)

    def test_tokens_property(self):
        prompt_section_base = TestablePromptSectionBase(
            tokens=0.5, required=True, separator=";", text_prefix="prefix"
        )
        self.assertEqual(prompt_section_base.tokens, 0.5)

    def test_separator_property(self):
        prompt_section_base = TestablePromptSectionBase(
            tokens=0.5, required=True, separator=";", text_prefix="prefix"
        )
        self.assertEqual(prompt_section_base.separator, ";")

    def test_text_prefix_property(self):
        prompt_section_base = TestablePromptSectionBase(
            tokens=0.5, required=True, separator=";", text_prefix="prefix"
        )
        self.assertEqual(prompt_section_base.text_prefix, "prefix")

    def test_required_property_default(self):
        prompt_section_base = TestablePromptSectionBase()
        self.assertEqual(prompt_section_base.required, True)

    def test_tokens_property_default(self):
        prompt_section_base = TestablePromptSectionBase()
        self.assertEqual(prompt_section_base.tokens, -1)

    def test_separator_property_default(self):
        prompt_section_base = TestablePromptSectionBase()
        self.assertEqual(prompt_section_base.separator, "\n")

    def test_text_prefix_property_default(self):
        prompt_section_base = TestablePromptSectionBase()
        self.assertEqual(prompt_section_base.text_prefix, "")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class AzureOpenAIEmbeddingsOptions:
    """
    Options for configuring an `AzureOpenAIEmbeddings` to generate embeddings.
    """

    azure_api_key: str
    "API key to use when making requests to Azure OpenAI."

    azure_endpoint: str
    "Deployment endpoint to use."

    azure_deployment: str
    "Name of the Azure OpenAI deployment (model) to use."

    azure_api_version: Optional[str] = None
    "Optional. Version of the API being called. Defaults to `2023-05-15`."

    log_requests: Optional[bool] = False
    "Whether to log requests to the console, useful for debugging and defaults to `false`"

    retry_policy: Optional[List[int]] = None
    "Optional. Retry policy to use in seconds. The default retry policy is `[2, 5]`."

    request_config: Optional[Dict[str, str]] = None
    "Request options to use."



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section_base import PromptSectionBase


class TextSection(PromptSectionBase):
    """
    A section of text that will be rendered as a message.

    Attributes:
        text (str): Text to use for this section.
        role (str): Message role to use for this section.

    """

    _text: str
    _role: str
    _length: int

    def __init__(
        self,
        text: str,
        role: str,
        tokens: float = -1,
        required: bool = True,
        separator: str = "\n",
        text_prefix: str = "",
    ):
        """
        Creates a new 'TextSection' instance.

        Args:
            text (str): Text to use for this section.
            role (str): Message role to use for this section.
            tokens (float, optional): Sizing strategy for this section. Defaults to `auto`.
            required (bool, optional): Indicates if this section is required. Defaults to `true`.
            separator (str, optional): Separator to use between sections when rendering as text.
              Defaults to `\n`.
            text_prefix (str, optional): Prefix to use for text output. Defaults to ``.

        """
        super().__init__(tokens, required, separator, text_prefix)
        self._text = text
        self._role = role
        self._length = -1

    @property
    def text(self):
        """Text to use for this section."""
        return self._text

    @property
    def role(self):
        """Message role to use for this section."""
        return self._role

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        # Calculate and cache length
        if self._length < 0:
            self._length = len(tokenizer.encode(self.text))

        # Return output
        messages: List[Message] = [Message(self.role, self.text)] if self._length > 0 else []
        return self._return_messages(messages, self._length, tokenizer, max_tokens)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional

from .augmentation_config import AugmentationConfig
from .completion_config import CompletionConfig


@dataclass
class PromptTemplateConfig:
    """
    Serialized prompt template configuration.

    Attributes:
        schema (float): The schema version of the prompt template. Can be '1' or '1.1'.

        completion (CompletionConfig): Completion settings for the prompt.

        type (Literal['completion']): Type of prompt template. Should always be 'completion'.

        augmentation (Optional[AugmentationConfig]): Augmentation settings for the prompt.
          New in schema version 1.1.

        description (Optional[str]): Description of the prompts purpose.

        default_backends (Optional[List[str]]): Array of backends (models) to use for the prompt.
          Passing the name of a model to use here will override the default model used by a planner.
          Deprecated: Use `completion.model` instead.

        augmentation (Optional[AugmentationConfig]): Optional augmentation settings for the prompt.
          New in schema version 1.1
    """

    schema: float
    completion: CompletionConfig
    type: Literal["completion"] = "completion"
    description: Optional[str] = None
    default_backends: Optional[List[str]] = None
    augmentation: Optional[AugmentationConfig] = None

    @classmethod
    def from_dict(cls, data: dict) -> "PromptTemplateConfig":
        config = cls(
            schema=data.get("schema", 1.0),
            completion=CompletionConfig.from_dict(data["completion"]),
            type=data.get("type", "completion"),
            description=data.get("description"),
            default_backends=data.get("default_backends"),
        )

        if data.get("augmentation"):
            config.augmentation = AugmentationConfig.from_dict(
                data.get("augmentation")  # type: ignore[arg-type]
            )

        return config



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.models import ChatCompletionAction, PromptResponse
from teams.ai.prompts import FunctionCall, Message
from teams.ai.tokenizers import GPTTokenizer
from teams.ai.validators import ActionResponseValidator, ValidatedChatCompletionAction
from teams.state import TurnState


class TestActionResponseValidator(IsolatedAsyncioTestCase):
    async def test_should_be_valid_when_no_message(self):
        validator = ActionResponseValidator(actions=[])
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(),
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)

    async def test_should_be_invalid_when_no_message_but_is_required(self):
        validator = ActionResponseValidator(actions=[], required=True)
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(
            res.feedback, "No action was specified. Call a action with valid arguments."
        )

    async def test_should_be_valid_when_no_function(self):
        validator = ActionResponseValidator(actions=[])
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=Message(role="assistant")),
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)

    async def test_should_be_invalid_when_no_function_but_is_required(self):
        validator = ActionResponseValidator(actions=[], required=True)
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=Message(role="assistant")),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(
            res.feedback, "No action was specified. Call a action with valid arguments."
        )

    async def test_should_be_invalid_when_no_function_name(self):
        validator = ActionResponseValidator(actions=[], required=True)
        message = Message(role="assistant", function_call=FunctionCall(name=None, arguments=None))

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=message),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, "action name missing. Specify a valid action name.")

    async def test_should_be_invalid_when_action_not_found(self):
        validator = ActionResponseValidator(actions=[ChatCompletionAction(name="a")], required=True)

        message = Message(role="assistant", function_call=FunctionCall(name="test", arguments=None))

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=message),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, 'Unknown action named "test". Specify a valid action name.')

    async def test_should_be_invalid_when_action_params_invalid(self):
        validator = ActionResponseValidator(
            actions=[
                ChatCompletionAction(
                    name="test",
                    parameters={
                        "type": "object",
                        "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
                        "required": ["name", "age"],
                    },
                )
            ],
            required=True,
        )

        message = Message(
            role="assistant",
            function_call=FunctionCall(name="test", arguments='{ "name": "test", "age": "10" }'),
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=message),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(
            res.feedback,
            'The action arguments had errors. Apply these fixes and call "test" '
            "action again:\n'10' is not of type 'number'",
        )

    async def test_should_be_valid(self):
        validator = ActionResponseValidator(
            actions=[
                ChatCompletionAction(
                    name="test",
                    parameters={
                        "type": "object",
                        "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
                        "required": ["name", "age"],
                    },
                )
            ],
            required=True,
        )

        message = Message(
            role="assistant",
            function_call=FunctionCall(name="test", arguments='{ "name": "test", "age": 10 }'),
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(message=message),
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)
        self.assertEqual(res.feedback, None)
        self.assertEqual(
            res.value,
            ValidatedChatCompletionAction(name="test", parameters={"name": "test", "age": 10}),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import re
from typing import (
    Awaitable,
    Callable,
    Generic,
    List,
    Optional,
    Pattern,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from aiohttp.web import Request, Response
from botbuilder.core import Bot, TurnContext
from botbuilder.schema import Activity, ActivityTypes, InvokeResponse
from botbuilder.schema.teams import (
    FileConsentCardResponse,
    O365ConnectorCardActionQuery,
)

from .activity_type import (
    ActivityType,
    ConversationUpdateType,
    MessageReactionType,
    MessageUpdateType,
)
from .adaptive_cards.adaptive_cards import AdaptiveCards
from .ai import AI
from .app_error import ApplicationError
from .app_options import ApplicationOptions
from .auth import AuthManager, OAuth, OAuthOptions
from .meetings.meetings import Meetings
from .message_extensions.message_extensions import MessageExtensions
from .route import Route, RouteHandler
from .state import TurnState
from .task_modules import TaskModules
from .teams_adapter import TeamsAdapter
from .typing import Typing

StateT = TypeVar("StateT", bound=TurnState)
IN_SIGN_IN_KEY = "__InSignInFlow__"


class Application(Bot, Generic[StateT]):
    """
    Application class for routing and processing incoming requests.

    The Application object replaces the traditional ActivityHandler that
    a bot would use. It supports a simpler fluent style of authoring bots
    versus the inheritance based approach used by the ActivityHandler class.

    Additionally, it has built-in support for calling into the SDK's AI system
    and can be used to create bots that leverage Large Language Models (LLM)
    and other AI capabilities.
    """

    typing: Typing

    _ai: Optional[AI[StateT]]
    _adaptive_card: AdaptiveCards[StateT]
    _options: ApplicationOptions
    _adapter: Optional[TeamsAdapter] = None
    _auth: Optional[AuthManager[StateT]] = None
    _before_turn: List[RouteHandler[StateT]] = []
    _after_turn: List[RouteHandler[StateT]] = []
    _routes: List[Route[StateT]] = []
    _error: Optional[Callable[[TurnContext, Exception], Awaitable[None]]] = None
    _turn_state_factory: Optional[Callable[[TurnContext], Awaitable[StateT]]] = None
    _message_extensions: MessageExtensions[StateT]
    _task_modules: TaskModules[StateT]
    _meetings: Meetings[StateT]

    def __init__(self, options: ApplicationOptions = ApplicationOptions()) -> None:
        """
        Creates a new Application instance.
        """
        self.typing = Typing()
        self._ai = AI[StateT](options.ai, logger=options.logger) if options.ai else None
        self._options = options
        self._routes = []
        self._message_extensions = MessageExtensions[StateT](self._routes)
        self._adaptive_card = AdaptiveCards[StateT](
            self._routes, options.adaptive_cards.action_submit_filer
        )
        self._task_modules = TaskModules[StateT](
            self._routes, options.task_modules.task_data_filter
        )
        self._meetings = Meetings[StateT](self._routes)

        if options.long_running_messages and (not options.adapter or not options.bot_app_id):
            raise ApplicationError("""
                The `ApplicationOptions.long_running_messages` property is unavailable because 
                no adapter or `bot_app_id` was configured.
                """)

        if options.adapter:
            self._adapter = options.adapter

        if options.auth:
            self._auth = AuthManager[StateT](default=options.auth.default)

            for name, opts in options.auth.settings.items():
                if isinstance(opts, OAuthOptions):
                    self._auth.set(name, OAuth[StateT](opts))

    @property
    def adapter(self) -> TeamsAdapter:
        """
        The bot's adapter.
        """

        if not self._adapter:
            raise ApplicationError("""
                The Application.adapter property is unavailable because it was 
                not configured when creating the Application.
                """)

        return self._adapter

    @property
    def ai(self) -> AI[StateT]:
        """
        This property is only available if the Application was configured with 'ai' options.
        An exception will be thrown if you attempt to access it otherwise.
        """

        if not self._ai:
            raise ApplicationError("""
                The `Application.ai` property is unavailable because no AI options were configured.
                """)

        return self._ai

    @property
    def auth(self) -> AuthManager[StateT]:
        """
        The application's authentication manager
        """
        if not self._auth:
            raise ApplicationError("""
                The `Application.auth` property is unavailable because
                no Auth options were configured.
                """)

        return self._auth

    @property
    def options(self) -> ApplicationOptions:
        """
        The application's configured options.
        """
        return self._options

    @property
    def message_extensions(self) -> MessageExtensions[StateT]:
        """
        Message Extensions
        """
        return self._message_extensions

    @property
    def adaptive_cards(self) -> AdaptiveCards[StateT]:
        """
        Access the application's adaptive cards functionalities.
        """
        return self._adaptive_card

    @property
    def task_modules(self) -> TaskModules[StateT]:
        """
        Access the application's task modules functionalities.
        """
        return self._task_modules

    @property
    def meetings(self) -> Meetings[StateT]:
        """
        Access the application's meetings functionalities.
        """
        return self._meetings

    def activity(
        self, type: ActivityType
    ) -> Callable[[RouteHandler[StateT]], RouteHandler[StateT]]:
        """
        Registers a new activity event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.activity("event")
        async def on_event(context: TurnContext, state: TurnState):
            print("hello world!")
            return True

        # Pass a function to this method
        app.action("event")(on_event)
        ```

        #### Args:
        - `type`: The type of the activity
        """

        def __selector__(context: TurnContext):
            return type == str(context.activity.type)

        def __call__(func: RouteHandler[StateT]) -> RouteHandler[StateT]:
            self._routes.append(Route[StateT](__selector__, func))
            return func

        return __call__

    def message(
        self, select: Union[str, Pattern[str]]
    ) -> Callable[[RouteHandler[StateT]], RouteHandler[StateT]]:
        """
        Registers a new message activity event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.message("hi")
        async def on_hi_message(context: TurnContext, state: TurnState):
            print("hello!")
            return True

        # Pass a function to this method
        app.message("hi")(on_hi_message)
        ```

        #### Args:
        - `select`: a string or regex pattern
        """

        def __selector__(context: TurnContext):
            if context.activity.type != ActivityTypes.message:
                return False

            if isinstance(select, Pattern):
                text = context.activity.text if context.activity.text else ""
                hits = re.match(select, text)
                return hits is not None

            i = context.activity.text.find(select)
            return i > -1

        def __call__(func: RouteHandler[StateT]) -> RouteHandler[StateT]:
            self._routes.append(Route[StateT](__selector__, func))
            return func

        return __call__

    def conversation_update(
        self, type: ConversationUpdateType
    ) -> Callable[[RouteHandler[StateT]], RouteHandler[StateT]]:
        """
        Registers a new message activity event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.conversation_update("channelCreated")
        async def on_channel_created(context: TurnContext, state: TurnState):
            print("a new channel was created!")
            return True

        # Pass a function to this method
        app.conversation_update("channelCreated")(on_channel_created)
        ```

        #### Args:
        - `type`: a string or regex pattern
        """

        def __selector__(context: TurnContext):
            if context.activity.type != ActivityTypes.conversation_update:
                return False

            if type == "membersAdded":
                if isinstance(context.activity.members_added, List):
                    return len(context.activity.members_added) > 0
                return False

            if type == "membersRemoved":
                if isinstance(context.activity.members_removed, List):
                    return len(context.activity.members_removed) > 0
                return False

            if isinstance(context.activity.channel_data, object):
                data = vars(context.activity.channel_data)
                return data["event_type"] == type

            return False

        def __call__(func: RouteHandler[StateT]) -> RouteHandler[StateT]:
            self._routes.append(Route[StateT](__selector__, func))
            return func

        return __call__

    def message_reaction(
        self, type: MessageReactionType
    ) -> Callable[[RouteHandler[StateT]], RouteHandler[StateT]]:
        """
        Registers a new message activity event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.message_reaction("reactionsAdded")
        async def on_reactions_added(context: TurnContext, state: TurnState):
            print("reactions was added!")
            return True

        # Pass a function to this method
        app.message_reaction("reactionsAdded")(on_reactions_added)
        ```

        #### Args:
        - `type`: a string or regex pattern
        """

        def __selector__(context: TurnContext):
            if context.activity.type != ActivityTypes.message_reaction:
                return False

            if type == "reactionsAdded":
                if isinstance(context.activity.reactions_added, List):
                    return len(context.activity.reactions_added) > 0
                return False

            if type == "reactionsRemoved":
                if isinstance(context.activity.reactions_removed, List):
                    return len(context.activity.reactions_removed) > 0
                return False

            return False

        def __call__(func: RouteHandler[StateT]) -> RouteHandler[StateT]:
            self._routes.append(Route[StateT](__selector__, func))
            return func

        return __call__

    def message_update(
        self, type: MessageUpdateType
    ) -> Callable[[RouteHandler[StateT]], RouteHandler[StateT]]:
        """
        Registers a new message activity event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.message_update("editMessage")
        async def on_edit_message(context: TurnContext, state: TurnState):
            print("message was edited!")
            return True

        # Pass a function to this method
        app.message_update("editMessage")(on_edit_message)
        ```

        #### Args:
        - `type`: a string or regex pattern
        """

        def __selector__(context: TurnContext):
            if type == "editMessage":
                if context.activity.type == ActivityTypes.message_update and isinstance(
                    context.activity.channel_data, dict
                ):
                    data = context.activity.channel_data
                    return data["event_type"] == type
                return False

            if type == "softDeleteMessage":
                if context.activity.type == ActivityTypes.message_delete and isinstance(
                    context.activity.channel_data, dict
                ):
                    data = context.activity.channel_data
                    return data["event_type"] == type
                return False

            if type == "undeleteMessage":
                if context.activity.type == ActivityTypes.message_update and isinstance(
                    context.activity.channel_data, dict
                ):
                    data = context.activity.channel_data
                    return data["event_type"] == type
                return False
            return False

        def __call__(func: RouteHandler[StateT]) -> RouteHandler[StateT]:
            self._routes.append(Route[StateT](__selector__, func))
            return func

        return __call__

    def file_consent_accept(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]]],
        Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]],
    ]:
        """
        Registers a handler for when a file consent card is accepted by the user.
         ```python
        # Use this method as a decorator
        @app.file_consent_accept()
        async def on_file_consent_accept(
            context: TurnContext, state: TurnState, file_consent_response: FileConsentCardResponse
        ):
            print(file_consent_response)
        # Pass a function to this method
        app.file_consent_accept()(on_file_consent_accept)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == "fileConsent/invoke"
                and isinstance(context.activity.value, dict)
                and context.activity.value.get("action") == "accept"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value)
                return True

            self._routes.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    def file_consent_decline(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]]],
        Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]],
    ]:
        """
        Registers a handler for when a file consent card is declined by the user.
         ```python
        # Use this method as a decorator
        @app.file_consent_decline()
        async def on_file_consent_decline(
            context: TurnContext, state: TurnState, file_consent_response: FileConsentCardResponse
        ):
            print(file_consent_response)
        # Pass a function to this method
        app.file_consent_decline()(on_file_consent_decline)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == "fileConsent/invoke"
                and isinstance(context.activity.value, dict)
                and context.activity.value.get("action") == "decline"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, FileConsentCardResponse], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value)
                return True

            self._routes.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    def o365_connector_card_action(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, O365ConnectorCardActionQuery], Awaitable[None]]],
        Callable[[TurnContext, StateT, O365ConnectorCardActionQuery], Awaitable[None]],
    ]:
        """
        Registers a handler for when a O365 connector card action is received from the user.
         ```python
        # Use this method as a decorator
        @app.o365_connector_card_action()
        async def on_o365_connector_card_action(
            context: TurnContext, state: TurnState, query: O365ConnectorCardActionQuery
        ):
            print(query)
        # Pass a function to this method
        app.o365_connector_card_action()(on_o365_connector_card_action)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == "actionableMessage/executeAction"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, O365ConnectorCardActionQuery], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, O365ConnectorCardActionQuery], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value)
                return True

            self._routes.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    def handoff(
        self,
    ) -> Callable[
        [Callable[[TurnContext, StateT, str], Awaitable[None]]],
        Callable[[TurnContext, StateT, str], Awaitable[None]],
    ]:
        """
        Registers a handler to handoff conversations from one copilot to another.
         ```python
        # Use this method as a decorator
        @app.handoff
        async def on_handoff(
            context: TurnContext, state: TurnState, continuation: str
        ):
            print(query)
        # Pass a function to this method
        app.handoff()(on_handoff)
        ```
        """

        def __selector__(context: TurnContext) -> bool:
            return (
                context.activity.type == ActivityTypes.invoke
                and context.activity.name == "handoff/action"
            )

        def __call__(
            func: Callable[[TurnContext, StateT, str], Awaitable[None]],
        ) -> Callable[[TurnContext, StateT, str], Awaitable[None]]:
            async def __handler__(context: TurnContext, state: StateT):
                if not context.activity.value:
                    return False
                await func(context, state, context.activity.value["continuation"])
                await context.send_activity(
                    Activity(type=ActivityTypes.invoke_response, value=InvokeResponse(status=200))
                )
                return True

            self._routes.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    def before_turn(self, func: RouteHandler[StateT]) -> RouteHandler[StateT]:
        """
        Registers a new event listener that will be executed before turns.
        This method can be used as either a decorator or a method and
        is called in the order they are registered.

        ```python
        # Use this method as a decorator
        @app.before_turn
        async def on_before_turn(context: TurnContext, state: TurnState):
            print("hello world!")
            return True

        # Pass a function to this method
        app.before_turn(on_before_turn)
        ```
        """

        self._before_turn.append(func)
        return func

    def after_turn(self, func: RouteHandler[StateT]) -> RouteHandler[StateT]:
        """
        Registers a new event listener that will be executed after turns.
        This method can be used as either a decorator or a method and
        is called in the order they are registered.

        ```python
        # Use this method as a decorator
        @app.after_turn
        async def on_after_turn(context: TurnContext, state: TurnState):
            print("hello world!")
            return True

        # Pass a function to this method
        app.after_turn(on_after_turn)
        ```
        """

        self._after_turn.append(func)
        return func

    def error(
        self, func: Callable[[TurnContext, Exception], Awaitable[None]]
    ) -> Callable[[TurnContext, Exception], Awaitable[None]]:
        """
        Registers an error handler that will be called anytime
        the app throws an Exception

        ```python
        # Use this method as a decorator
        @app.error
        async def on_error(context: TurnContext, err: Exception):
            print(err.message)

        # Pass a function to this method
        app.error(on_error)
        ```
        """

        self._error = func

        if self._adapter:
            self._adapter.on_turn_error = func

        return func

    def turn_state_factory(self, func: Callable[[TurnContext], Awaitable[StateT]]):
        """
        Custom Turn State Factory
        """

        self._turn_state_factory = func
        return func

    async def process(self, request: Request) -> Optional[Response]:
        """
        Creates a turn context and runs the middleware pipeline for an incoming activity.
        """

        if not self._adapter:
            raise ApplicationError(
                "cannot call `app.process_activity` when `ApplicationOptions.adapter` not provided"
            )

        return await self._adapter.process(request, self)

    async def on_turn(self, context: TurnContext):
        await self._start_long_running_call(context, self._on_turn)

    async def _on_turn(self, context: TurnContext):
        try:
            if self._options.start_typing_timer:
                await self.typing.start(context)

            # remove @mentions
            if self.options.remove_recipient_mention:
                if context.activity.type == ActivityTypes.message:
                    context.activity.text = context.remove_recipient_mention(context.activity)

            state: StateT = cast(StateT, await TurnState.load(context, self._options.storage))

            if self._turn_state_factory:
                state = await self._turn_state_factory(context)

            await state.load(context, self._options.storage)
            state.temp.input = context.activity.text

            # authentication
            if (
                self.options.auth
                and self._auth
                and (
                    (
                        (
                            isinstance(self.options.auth.auto, bool)
                            and self.options.auth.auto is True
                        )
                        or (
                            callable(self.options.auth.auto)
                            and self.options.auth.auto(context) is True
                        )
                    )
                    or IN_SIGN_IN_KEY in state.user
                )
            ):
                key: Optional[str] = (
                    state.user[IN_SIGN_IN_KEY]
                    if IN_SIGN_IN_KEY in state.user
                    else self.options.auth.default
                )

                if key is not None:
                    state.user[IN_SIGN_IN_KEY] = key
                    res = await self._auth.sign_in(context, state, key=key)

                    if res.status == "complete":
                        del state.user[IN_SIGN_IN_KEY]

                    if res.status == "pending":
                        await state.save(context, self._options.storage)
                        return

                    if res.status == "error" and res.reason != "invalid-activity":
                        del state.user[IN_SIGN_IN_KEY]
                        raise ApplicationError(f"[{res.reason}] => {res.message}")

            # run before turn middleware
            for before_turn in self._before_turn:
                is_ok = await before_turn(context, state)

                if not is_ok:
                    await state.save(context, self._options.storage)
                    return

            # download input files
            if (
                self._options.file_downloaders is not None
                and len(self._options.file_downloaders) > 0
            ):
                input_files = state.temp.input_files if state.temp.input_files is not None else []
                for file_downloader in self._options.file_downloaders:
                    files = await file_downloader.download_files(context)
                    input_files.append(files)
                state.temp.input_files = input_files

            # run activity handlers
            is_ok, matches = await self._on_activity(context, state)

            if not is_ok:
                await state.save(context, self._options.storage)
                return

            # only run chain when no activity handlers matched
            if (
                matches == 0
                and self._ai
                and self._options.ai
                and context.activity.type == ActivityTypes.message
                and context.activity.text
            ):
                is_ok = await self._ai.run(context, state)

                if not is_ok:
                    await state.save(context, self._options.storage)
                    return

            # run after turn middleware
            for after_turn in self._after_turn:
                is_ok = await after_turn(context, state)

                if not is_ok:
                    await state.save(context, self._options.storage)
                    return

            await state.save(context, self._options.storage)
        except ApplicationError as err:
            await self._on_error(context, err)
        finally:
            self.typing.stop()

    async def _on_activity(self, context: TurnContext, state: StateT) -> Tuple[bool, int]:
        matches = 0

        # ensure we handle invokes first
        routes = filter(lambda r: not r.is_invoke and r.selector(context), self._routes)
        invoke_routes = filter(lambda r: r.is_invoke and r.selector(context), self._routes)

        for route in invoke_routes:
            if route.selector(context):
                matches = matches + 1

                if not await route.handler(context, state):
                    return False, matches

        for route in routes:
            if route.selector(context):
                matches = matches + 1

                if not await route.handler(context, state):
                    return False, matches

        return True, matches

    async def _start_long_running_call(
        self, context: TurnContext, func: Callable[[TurnContext], Awaitable]
    ):
        if (
            self._adapter
            and ActivityTypes.message == context.activity.type
            and self._options.long_running_messages
        ):
            return await self._adapter.continue_conversation(
                reference=context.get_conversation_reference(context.activity),
                callback=func,
                bot_app_id=self.options.bot_app_id,
            )

        return await func(context)

    async def _on_error(self, context: TurnContext, err: ApplicationError) -> None:
        if self._error:
            return await self._error(context, err)

        self._options.logger.error(err)
        raise err



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .chat_completion_action import ChatCompletionAction
from .openai_model import AzureOpenAIModelOptions, OpenAIModel, OpenAIModelOptions
from .prompt_completion_model import PromptCompletionModel
from .prompt_response import PromptResponse, PromptResponseStatus

__all__ = [
    "ChatCompletionAction",
    "AzureOpenAIModelOptions",
    "OpenAIModel",
    "OpenAIModelOptions",
    "PromptCompletionModel",
    "PromptResponse",
    "PromptResponseStatus",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import UserDict
from typing import Any, Dict, MutableMapping, Optional, Tuple

from ..app_error import ApplicationError


class MemoryBase(ABC):
    """
    Memory Base
    """

    @abstractmethod
    def has(self, path: str) -> bool:
        """
        Checks if a value exists in the memory.

        Args:
            path (str): Path to the value to check in the form of `[scope].property`.
              If scope is omitted, the value is checked in the temporary scope.

        Returns:
            bool: True if the value exists, False otherwise.
        """

    @abstractmethod
    def get(self, path: str) -> Optional[Any]:
        """
        Retrieves a value from the memory.

        Args:
            path (str): Path to the value to retrieve in the form of `[scope].property`.
              If scope is omitted, the value is retrieved from the temporary scope.

        Returns:
            Any: The value or None if not found.
        """

    @abstractmethod
    def set(self, path: str, value: Any) -> None:
        """
        Assigns a value to the memory.

        Args:
            path (str): Path to the value to assign in the form of `[scope].property`.
              If scope is omitted, the value is assigned to the temporary scope.
            value (Any): Value to assign.
        """

    @abstractmethod
    def delete(self, path: str) -> None:
        """
        Deletes a value from the memory.

        Args:
            path (str): Path to the value to delete in the form of `[scope].property`.
              If scope is omitted, the value is deleted from the temporary scope.
        """

    def _get_scope_and_name(self, path: str) -> Tuple[str, str]:
        parts = path.split(".")

        if len(parts) > 2:
            raise ApplicationError(f"Invalid state path: {path}")

        if len(parts) == 1:
            parts.insert(0, "temp")

        return (parts[0], parts[1])


class Memory(MemoryBase):
    """
    Represents a memory.

    A memory is a key-value store that can be used to store and retrieve values.
    """

    _parent: Optional[MemoryBase]
    _scopes: Dict[str, MutableMapping[str, Any]]

    def __init__(self, parent: Optional[MemoryBase] = None) -> None:
        self._parent = parent
        self._scopes = {}

    def has(self, path: str) -> bool:
        scope, name = self._get_scope_and_name(path)

        if scope in self._scopes and name in self._scopes[scope]:
            return True

        if self._parent:
            return self._parent.has(path)

        return False

    def get(self, path: str) -> Optional[Any]:
        scope, name = self._get_scope_and_name(path)

        if scope in self._scopes and name in self._scopes[scope]:
            return self._scopes[scope][name]

        if self._parent:
            return self._parent.get(path)

        return None

    def set(self, path: str, value: Any) -> None:
        scope, name = self._get_scope_and_name(path)

        if not scope in self._scopes:
            self._scopes[scope] = UserDict()

        self._scopes[scope][name] = value

    def delete(self, path: str) -> None:
        scope, name = self._get_scope_and_name(path)

        if scope in self._scopes and name in self._scopes[scope]:
            del self._scopes[scope][name]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from .sections.template_section import TemplateSection


class AssistantMessage(TemplateSection):
    """
    A message sent by the assistant.
    """

    def __init__(self, template: str, tokens: float = -1, text_prefix: str = "assistant: "):
        """
        Creates a new 'AssistantMessage' instance.

        Args:
            template (str): Template to use for this section.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
            text_prefix (str, optional): Prefix to use for assistant messages when
              rendering as text. Defaults to `assistant: `.
        """
        super().__init__(template, "assistant", tokens, True, "\n", text_prefix)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from dataclasses_json import DataClassJsonMixin, dataclass_json


@dataclass_json
@dataclass
class ChatCompletionAction(DataClassJsonMixin):
    """
    An action that can be called by an LLM.
    """

    name: str
    """
    Name of the action to be called.
    """

    description: Optional[str] = None
    """
    Optional. Description of what the action does.
    """

    parameters: Optional[Dict[str, Any]] = None
    """
    Optional. Parameters the action accepts, described as a JSON Schema object.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from msrest.serialization import Model


@dataclass
class ClientCitation(Model):
    """
    Represents the citations payload to be sent to the client.

    Attributes:
        @type (str): Required; must be 'Claim'
        position (str): The number and position of the citation
        appearance (Appearance): The appearance object within Citation
    """

    _attribute_map = {
        "type_": {"key": "@type", "type": "str"},
    }

    type_: str = field(default="Claim", metadata={"alias": "@type"}, init=False, repr=False)
    position: str
    appearance: Appearance


@dataclass
class Appearance(Model):
    """
    Represents how the citation will be rendered

    Attributes:
        @type (str): Required; must be 'DigitalDocument'
        name (str): The name of the document
        text (str): Optional; ignored in Teams
        url (str): The url of the document
        abstract (str): Content of the citation. Should be clipped if longer than ~500 characters
        encodingFormat (str): The encoding format of the citation
        image (str): Used for icon; for not it is ignored
        keywords (list[str]): The optional keywords to the citation
        usageInfo (SensitivityUsageInfo): The optional sensitivity content information
    """

    _attribute_map = {
        "type_": {"key": "@type", "type": "str"},
        "usage_info": {"key": "usageInfo", "type": "SensitivityUsageInfo"},
        "encoding_format": {"key": "encodingFormat", "type": "str"},
    }

    name: str
    abstract: str
    usage_info: Optional[SensitivityUsageInfo] = field(
        default=None, metadata={"alias": "usageInfo"}, init=False, repr=False
    )
    keywords: Optional[list[str]] = None
    type_: str = field(default="DigitalDocument", metadata={"alias": "@type"})
    text: Optional[str] = ""
    url: str = ""
    encoding_format: Optional[str] = "text/html"
    image: Optional[str] = ""


@dataclass
class SensitivityUsageInfo(Model):
    """
    Represents the sensitivity usage info for content sent to the user.
    This is used to provide information about the content to the user.

    Attributes:
        type (str): Required; must be 'https://schema.org/Message'

    """

    _attribute_map = {
        "type_": {"key": "@type", "type": "str"},
    }

    name: str
    description: Optional[str]
    position: Optional[int]
    pattern: Optional[Pattern]
    type_: str = field(default="https://schema.org/Message", metadata={"alias": "@type"})


@dataclass
class Pattern(Model):
    """
    Attributes:
        type (str): Required; must be 'DefinedTerm'
        inDefinedTermSet (str)
        name (str): Color
        termCode (str): The color code e.g. #454545

    """

    _attribute_map = {
        "type_": {"key": "@type", "type": "str"},
        "in_defined_term_set": {"key": "inDefinedTermSet", "type": "str"},
        "term_code": {"key": "termCode", "type": "str"},
    }

    in_defined_term_set: str
    name: str
    term_code: str
    type_: str = field(default="DefinedTerm", metadata={"alias": "@type"})



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Optional

import teams.ai.augmentations

from ..models.chat_completion_action import ChatCompletionAction
from .prompt_template_config import PromptTemplateConfig
from .sections.prompt_section import PromptSection


@dataclass
class PromptTemplate:
    """
    Prompt template cached by the prompt manager.

    Attributes:
        name (str): Name of the prompt template.

        prompt (PromptSection): Text of the prompt template.

        config (PromptTemplateConfig): Configuration settings for the prompt template.

        actions (Optional[List[ChatCompletionAction]]): Optional
            list of actions the model may generate JSON inputs for.

        augmentation (Optional[Augmentation]): Optional augmentation for the prompt template.
    """

    name: str
    prompt: PromptSection
    config: PromptTemplateConfig
    actions: Optional[List[ChatCompletionAction]] = None
    augmentation: Optional[teams.ai.augmentations.Augmentation] = None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import List

from botbuilder.core import BotAdapter
from botbuilder.schema import Activity, ResourceResponse

from .activity import ACTIVITY


class SimpleAdapter(BotAdapter):
    async def send_activities(self, context, activities) -> List[ResourceResponse]:
        responses = []

        assert context is not None
        assert activities is not None
        assert isinstance(activities, list)

        for _, activity in enumerate(activities):
            assert isinstance(activity, Activity)
            responses.append(ResourceResponse(id="5678"))

        return responses

    async def update_activity(self, context, activity):
        assert context is not None
        assert activity is not None
        return ResourceResponse(id=activity.id)

    async def delete_activity(self, context, reference):
        assert context is not None
        assert reference is not None
        assert reference.activity_id == ACTIVITY.id



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from typing import Any, Optional


def parse_object(text: str) -> Optional[Any]:
    """
    Parse JSON Object
    """

    start_brace = text.find("{")

    if start_brace == -1:
        return None

    obj = text[start_brace:]
    nesting = ["}"]
    cleaned = "{"
    in_string = False
    i = 1

    while i < len(obj) and len(nesting) > 0:
        char = obj[i]

        if in_string:
            cleaned += char

            if char == "\\":
                while char == "\\":
                    i += 1
                    char = obj[i]

                if i < len(obj):
                    cleaned += obj[i]
                else:
                    return None
            elif char == '"':
                in_string = False
        else:
            add_pre = False
            add_post = False

            if char == '"':
                in_string = True
            elif char == "{":
                nesting.append("}")
            elif char == "[":
                nesting.append("]")
            elif char in ("}", "]"):
                close = nesting.pop()

                if close != char:
                    return None
            elif char == "<":
                add_pre = True
            elif char == ">":
                add_post = True

            if add_pre:
                cleaned += '"'

            cleaned += char

            if add_post:
                cleaned += '"'

        i += 1

    if len(nesting) > 0:
        nesting.reverse()
        cleaned += "".join(nesting)

    try:
        return json.loads(cleaned)
    except ValueError:
        return None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class AdaptiveCardsOptions:
    action_submit_filer: str = "verb"
    """
    Data field used to identify the Action.Submit handler to trigger.
    When an Action.Submit is triggered, the field name specified here will be used to determine
    the handler to route the request to.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import re
from typing import Awaitable, Callable, Generic, List, Pattern, TypeVar, Union

from botbuilder.core import TurnContext
from botbuilder.core.serializer_helper import serializer_helper
from botbuilder.schema import Activity, ActivityTypes, InvokeResponse
from botbuilder.schema.teams import (
    TaskModuleContinueResponse,
    TaskModuleMessageResponse,
    TaskModuleResponse,
    TaskModuleTaskInfo,
)

from teams.route import Route
from teams.state import TurnState

FETCH_INVOKE_NAME = "task/fetch"
SUBMIT_INVOKE_NAME = "task/submit"

StateT = TypeVar("StateT", bound=TurnState)


class TaskModules(Generic[StateT]):
    _route_registry: List[Route[StateT]]
    _task_data_filter: str

    def __init__(self, route_registry: List[Route[StateT]], task_data_filter: str) -> None:
        self._route_registry = route_registry
        self._task_data_filter = task_data_filter

    def fetch(self, verb: Union[str, Pattern[str], Callable[[TurnContext], bool]]) -> Callable[
        [Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str]]]],
        Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str]]],
    ]:
        """
        Adds a route for handling the initial fetch of the task module.
        This method can be used as either a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.task_modules.handle_fetch("buy")
        async def fetch(context: TurnContext, state: TurnState, data: dict):
            print(f"Execute with data: {data}")
            return True

        # Pass a function to this method
        app.adaptive_cards.fetch("buy")(handle_fetch)
        ```

        #### Args:
        - `verb`: a string, regex pattern or a function to match the verb of the request
        """

        # Create route selector for the handler
        def __selector__(context: TurnContext) -> bool:
            return self._task_modules_selector(
                context, verb, self._task_data_filter, FETCH_INVOKE_NAME
            )

        def __call__(
            func: Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str]]],
        ) -> Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str]]]:
            async def __handler__(context: TurnContext, state: StateT) -> bool:
                # the selector already ensures data exists
                result = await func(context, state, context.activity.value["data"])
                await self._send_response(context, result)
                return True

            self._route_registry.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    # TODO: Ported the code first. Need more discussion on the functionality.
    def submit(self, verb: Union[str, Pattern[str], Callable[[TurnContext], bool]]) -> Callable[
        [Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str, None]]]],
        Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str, None]]],
    ]:
        """
        Adds a route for handling the submission of a task module.
        This method can be used as either a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.task_modules.submit("confirm")
        async def handle_submit(context: TurnContext, state: TurnState, data: dict):
            print(f"Execute with data: {data}")
            return True

        # Pass a function to this method
        app.adaptive_cards.submit("confirm")(handle_submit)
        ```

        #### Args:
        - `verb`: a string, regex pattern or a function to match the verb of the request
        """

        # Create route selector for the handler
        def __selector__(context: TurnContext) -> bool:
            return self._task_modules_selector(
                context, verb, self._task_data_filter, SUBMIT_INVOKE_NAME
            )

        def __call__(
            func: Callable[
                [TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str, None]]
            ],
        ) -> Callable[[TurnContext, StateT, dict], Awaitable[Union[TaskModuleTaskInfo, str, None]]]:
            async def __handler__(context: TurnContext, state: StateT) -> bool:
                # the selector already ensures data exists
                result = await func(context, state, context.activity.value["data"])
                await self._send_response(context, result)
                return True

            self._route_registry.append(Route[StateT](__selector__, __handler__, True))
            return func

        return __call__

    def _task_modules_selector(
        self,
        context: TurnContext,
        verb: Union[str, Pattern[str], Callable[[TurnContext], bool]],
        filter_field: str,
        invoke_name: str,
    ) -> bool:
        if context.activity.type == ActivityTypes.invoke and context.activity.name == invoke_name:
            data = context.activity.value["data"] if context.activity.value else None
            if isinstance(data, dict) and isinstance(data[filter_field], str):
                # when verb is a function
                if callable(verb):
                    return verb(context)
                # when verb is a regex pattern
                if isinstance(verb, Pattern):
                    hits = re.match(verb, data[filter_field])
                    return hits is not None
                # when verb is a string
                return verb == data[filter_field]

        return False

    async def _send_response(self, context: TurnContext, result):
        if context.turn_state.get(context._INVOKE_RESPONSE_KEY) is None:
            if isinstance(result, str):
                response = TaskModuleResponse(task=TaskModuleMessageResponse(value=result))
            elif isinstance(result, TaskModuleTaskInfo):
                response = TaskModuleResponse(task=TaskModuleContinueResponse(value=result))
            else:
                response = None

            await context.send_activity(
                Activity(
                    type="invokeResponse",
                    value=InvokeResponse(status=200, body=serializer_helper(response)),
                )
            )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .gpt_tokenizer import GPTTokenizer
from .tokenizer import Tokenizer

__all__ = ["GPTTokenizer", "Tokenizer"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ...user_agent import _UserAgent
from ..prompts.prompt_functions import PromptFunctions
from ..prompts.prompt_template import PromptTemplate
from ..tokenizers import Tokenizer
from .prompt_response import PromptResponse


class PromptCompletionModel(ABC, _UserAgent):
    """
    An AI model that can be used to complete prompts.
    """

    @abstractmethod
    async def complete_prompt(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        template: PromptTemplate,
    ) -> PromptResponse[str]:
        """
        Completes a prompt.

        Args:
            context (TurnContext): turn context.
            memory (MemoryBase): the turn state.
            functions (PromptFunctions): to use when rendering the prompt.
            tokenizer (Tokenizer): to use when rendering the prompt.
            template (PromptTemplate): template to complete.

        Returns:
            PromptResponse[str]: a response with the status and message.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

ParamT = TypeVar("ParamT")


@dataclass
class Query(Generic[ParamT]):
    count: int
    skip: int
    parameters: ParamT



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, Generic, Optional, TypeVar, cast

from botbuilder.core import Storage, TurnContext

from .conversation_state import ConversationState
from .memory import MemoryBase
from .state import State, state
from .temp_state import TempState
from .user_state import UserState

ConversationStateT = TypeVar("ConversationStateT", bound=ConversationState)
UserStateT = TypeVar("UserStateT", bound=UserState)
TempStateT = TypeVar("TempStateT", bound=TempState)


@state
class TurnState(MemoryBase, State, Generic[ConversationStateT, UserStateT, TempStateT]):
    """
    Default Turn State
    """

    conversation: ConversationStateT
    user: UserStateT
    temp: TempStateT

    async def save(self, context: TurnContext, storage: Optional[Storage] = None) -> None:
        if storage and len(self.__deleted__) > 0:
            await storage.delete(self.__deleted__)

        for item in self.values():
            if isinstance(item, State):
                await item.save(context, storage)

    def has(self, path: str) -> bool:
        scope, name = self._get_scope_and_name(path)
        return scope in self and name in self[scope]

    def get(self, path: str) -> Optional[Any]:  # type: ignore[override]
        if not self.has(path):
            return None

        scope, name = self._get_scope_and_name(path)
        return self[scope][name]

    def set(self, path: str, value: Any) -> None:
        scope, name = self._get_scope_and_name(path)

        if not scope in self:
            raise KeyError(f"[{self.__class__.__name__}.set]: '{scope}' is not defined")

        self[scope][name] = value

    def delete(self, path: str) -> None:
        if not self.has(path):
            return

        scope, name = self._get_scope_and_name(path)
        del self[scope][name]

    @classmethod
    async def load(
        cls, context: TurnContext, storage: Optional[Storage] = None
    ) -> "TurnState[ConversationStateT, UserStateT, TempStateT]":
        conversation = await ConversationState.load(context, storage)
        user = await UserState.load(context, storage)
        temp = await TempState.load(context, storage)

        return cls(
            conversation=cast(ConversationStateT, conversation),
            user=cast(UserStateT, user),
            temp=cast(TempStateT, temp),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from botbuilder.core import TurnContext
from jsonschema import ValidationError, validate

from ...state import MemoryBase
from ...utils.json import parse
from ..models.prompt_response import PromptResponse
from ..tokenizers import Tokenizer
from .prompt_response_validator import PromptResponseValidator
from .validation import Validation

NEW_LINE = "\n"


class JSONResponseValidator(PromptResponseValidator):
    """
    Default response validator that always returns true.
    """

    _schema: Optional[Dict[str, Any]]
    _missing_json_feedback: str
    _error_feedback: str

    def __init__(
        self,
        schema: Optional[Dict[str, Any]] = None,
        missing_json_feedback: Optional[str] = None,
        error_feedback: Optional[str] = None,
    ) -> None:
        """
        Creates a new `JSONResponseValidator` instance.
        """

        super().__init__()
        self._schema = schema
        self._missing_json_feedback = (
            missing_json_feedback
            if missing_json_feedback is not None
            else "No valid JSON objects were found in the response. Return a valid JSON object."
        )
        self._error_feedback = (
            error_feedback
            if error_feedback is not None
            else "The JSON returned had errors. Apply these fixes:"
        )

    @property
    def schema(self):
        """
        JSON schema to validate the response against.
        """
        return self._schema

    @property
    def missing_json_feedback(self):
        """
        Feedback given when no JSON is returned.
        """
        return self._missing_json_feedback

    @property
    def error_feedback(self):
        """
        Feedback prefix given when schema errors are detected.
        """
        return self._error_feedback

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse[str],
        remaining_attempts: int,
    ) -> Validation:
        message = response.message

        if message is None:
            return Validation(valid=False)

        content = message.content if message.content is not None else ""
        parsed = parse(content)

        if len(parsed) == 0:
            return Validation(valid=False, feedback=self.missing_json_feedback)

        if self._schema is None:
            return Validation(value=parsed.pop())

        parsed.reverse()
        errors: List[str] = []

        for obj in parsed:
            try:
                validate(instance=obj, schema=self._schema)
                return Validation(value=obj)
            except ValidationError as err:
                errors.append(err.message)

        return Validation(
            valid=False, feedback=f"{self._error_feedback}{NEW_LINE}{NEW_LINE.join(errors)}"
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.utils.citations import format_citations_response


class TestFormatCitationsResponse(TestCase):
    def test_format_citations_response(self):
        result = format_citations_response("hello [doc1] world [docs2]")
        self.assertEqual(result, "hello [1] world [2]")

    def test_format_citations_response_none(self):
        result = format_citations_response("hello world")
        self.assertEqual(result, "hello world")
